from pydantic import BaseModel, DirectoryPath, field_validator, model_validator, NewPath
from .patterns import Pattern
from typing import Optional

class ConfigError(Exception):
    def __init__(self, msg):
        message_primer = "There appears to be a problem with the values entered:\n"
        self.message = message_primer + msg
        super().__init__(self.message)

class ConfigStrategy(BaseModel):
    processed_directory: DirectoryPath
    move_to_directory: Optional[DirectoryPath] = None
    file_name_pattern: str
    file_content_pattern: Optional[str] = None
    file_format: str
    terms: dict[str, str]
    export_format: str
    export_path: NewPath
    export_csv_divider: Optional[str] = ';'
    # terms_patterns_group is created from 'terms', see @model_validator
    terms_patterns_group: dict[str, Pattern] = None

    @field_validator('file_name_pattern', 'export_format')
    @classmethod
    def non_empty_string(cls, v: str):
        assert v != '', 'Must be a non-empty string'
        return v.strip()

    @field_validator('file_name_pattern', 'file_content_pattern')
    @classmethod
    def selection_must_be_regex(cls, v: str):
        v = v.strip()
        if not Pattern.check_regex(v): 
            raise ConfigError(f"Pattern '{v}' cannot be used as a regex pattern; also, regex groups are not allowed!")
        return v
    
    @field_validator('export_format')
    @classmethod
    def validate_export_format(cls, ef: str):
        valid_formats = {"csv", "html", "json", "xlsx"}
        if ef not in valid_formats:
            raise ConfigError(msg=f"Concerning '{ef}': Export format must conform to one of these options: {valid_formats}!")
        return ef
    
    @field_validator('file_format')
    @classmethod
    def validate_file_format(cls, ff: str):
        valid_formats = {"txt", "pdf", "docx"}
        if ff not in valid_formats:
            raise ConfigError(msg=f"Concerning '{ff}': File format must conform to one of these options: {valid_formats}!")
        return ff
    
    @model_validator(mode='after')
    def check_terms_and_patterns(self):
        # process terms_and_patterns 
        processed_tp = {term:Pattern(patterns_and_options_string=pos) for term, pos in self.terms.items()}
        self.terms_patterns_group = processed_tp
        return self

class Config(BaseModel):
    title: str
    strategies: dict[str, ConfigStrategy]
    


import re
from typing import Any
from pydantic_core import CoreSchema, core_schema
from pydantic import GetCoreSchemaHandler

class PatternOptionsError(Exception):
    def __init__(self, msg):
        message_primer = "There appears to be a problem with the pattern options specified:"
        self.message = message_primer + msg
        super().__init__(self.message)
        
class PatternStringError(Exception):
    def __init__(self, msg):
        message_primer = "There appears to be a problem with the patterstring:\n"
        self.message = message_primer + msg
        super().__init__(self.message)

class PatternOptions:
    multiLine: bool = False
    cropToLength: int = None
    lineBreakToSpace: bool = False
    required: bool = True
    
    def __init__(self, optionstring: str):
        self.process_optionstring(optionstring)
    
    def process_optionstring(self, ops: str):
        # cropToLength specified?
        cropToLengthFinds = re.findall("c\(\d+\)", ops)
        if len(cropToLengthFinds) > 0:
            # extract number, only take first, no matter what
            numberString = re.findall("\d+", cropToLengthFinds[0])
            if len(numberString) > 0:
                try:
                    self.cropToLength = int(numberString[0])
                except:
                    pass
        # multiline specified?
        multilineFinds = re.findall("m", ops)
        if len(multilineFinds) > 0:
            self.multiLine = True
        # lineBreakToSpace specified?
        lineBreakToSpaceFinds = re.findall("l", ops)
        if len(lineBreakToSpaceFinds) > 0:
            self.lineBreakToSpace = True
        # not required specified?
        notRequiredFinds = re.findall("\?", ops)
        if len(notRequiredFinds) > 0:
            self.required = False
            
    def __repr__(self):
        return f"PatternOptions(cropToLength:{self.cropToLength}, multiLine:{self.multiLine}, lineBreakToSpace:{self.lineBreakToSpace}, required:{self.required})"

class Pattern:
    compiled_regex_pattern: re.Pattern = None
    pattern_options: PatternOptions = None
    patterns_and_options_string: str = None
    pattern_string: str = None
    options_string: str =  None
    pattern_divider: str = "@@@"
    options_divider: str = "==="
    group_of_interest: int = None
    
    def __repr__(self):
        return f"Pattern(compiled_regex_pattern:{self.compiled_regex_pattern},\
                pattern_options:{self.pattern_options},\
                patterns_and_options_string:{self.patterns_and_options_string},\
                options_string:{self.options_string},\
                group_of_interest:{self.group_of_interest})"
    
    # the following class method is necessary for pydantic to use this class as a type.
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, handler: GetCoreSchemaHandler) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))
    
    def __init__(self, patterns_and_options_string: str):
        self.patterns_and_options_string = patterns_and_options_string
        self.process_options()
        self.process_terms()
    
    # basic validation
    @classmethod
    def check_regex(cls, ps: str) -> bool:
        try: 
            rgx = re.compile(ps)
            # Also, do not allow regex groups
            if rgx.groups > 0:
                return False
        except:
            return False
        return True
    
    # helper to check if pattern only consists of a matchall pattern
    def matchall_only(self, s) -> bool:
        return re.search("\.\*", s) and len(s) == 2
    
    # extract options
    def process_options(self):
        # check number of dividers
        divider_hits = len(re.findall(self.options_divider, self.patterns_and_options_string))
        # check the number of occurrences of divider
        if divider_hits > 1:
            # as this is not implemented, throw an error
            raise PatternStringError(msg=f"Multiple option groups detected in '{self.patterns_and_options_string}'! Only one allowed!")
        # if no options present, set default options!
        if divider_hits == 0:
            self.pattern_options = PatternOptions("")
            self.pattern_string = self.patterns_and_options_string
            return
        # if options present, create new PatternOptions object
        options_split = re.split(self.options_divider, self.patterns_and_options_string)
        self.pattern_string = options_split[0]
        self.pattern_options = PatternOptions(options_split[1])
        
    # process_terms
    def process_terms(self):
        # extract rflag
        rflag = re.DOTALL if self.pattern_options.multiLine else re.NOFLAG
        # if patternstrings contains groups, reject
        if not self.check_regex(self.pattern_string):
            raise PatternStringError(f"Concerning '{self.pattern_string}': cannot be used as regex pattern; also, regex groups are not allowed!")
        # check if matchall pattern is present (as this is not allowed)
        if self.matchall_only(self.pattern_string):
            raise PatternStringError(msg=f"The string '{self.pattern_string}' only contains the matchall-pattern '.*' and can therefore not be processed.")
        # divider_hits counts the number of divider in the string; only one is allowed (see below)
        divider_hits = len(re.findall(self.pattern_divider, self.pattern_string))
        # check the number of occurrences of divider
        if divider_hits > 1:
            # as this is not implemented, throw an error
            raise PatternStringError(msg=f"Each term must correspond to either *one regex pattern* or *two regex patterns divided by '{divider}'*!")
        if divider_hits == 0:
            # return without capture groups
            self.compiled_regex_pattern = re.compile(self.pattern_string, rflag)
            return
            # return (re.compile(patternstring, rflag), -1, -1)
        # process the patternstrings divided by divider
        multiple_patternstrings = re.split(self.pattern_divider, self.pattern_string)
        
        # check if patternstring and multiple_patternstrings are valid regex patterns without groups
        all_strings = multiple_patternstrings.copy()
        all_strings.append(self.pattern_string)
        all_check = [self.check_regex(s) for s in all_strings]
        if not all(all_check):
            raise PatternStringError(f"Concerning one of '{all_strings!r}': cannot be used as regex pattern; also regex groups are not allowed!")
        
        # do any of the patternstrings only contain a matchall pattern?
        if any([self.matchall_only(p) for p in multiple_patternstrings]):
            raise PatternStringError(msg=f"At least one of '{multiple_patternstrings!r}' only consists of a matchall-pattern '.*' and can therefore not be processed.")
        # is any of the patternstrings of length 0? (This occurs if the matchall-pattern not surrounded by two patterns)
        lenx = [len(i) for i in multiple_patternstrings]
        lenx0 = [l == 0 for l in lenx]
        #print(f"multiple patternstrings: {multiple_patternstrings}, lenx: {lenx}, lenx0: {lenx0}\n")
        # if yes, create two groups
        if any(lenx0):
            # the first group is of interest
            if lenx0[0]:
                ps = multiple_patternstrings[1]
                self.compiled_regex_pattern = re.compile(f"(.*)({ps})", rflag)
                #print(f"ps: {ps}, rgx: {self.compiled_regex_pattern}\n")
                self.group_of_interest = 1
                # return (re.compile(f"(.*)({multiple_patternstrings[1]})", rflag), 1, 2)
            else:
            # the second group is of interest
                ps = multiple_patternstrings[0]
                self.compiled_regex_pattern = re.compile(f"({ps})(.*)", rflag)
                #print(f"ps: {ps}, rgx: {self.compiled_regex_pattern}\n")
                self.group_of_interest = 2
                # return (re.compile(f"({multiple_patternstrings[0]})(.*)", rflag), 2, 2)
        else:
            # none of the patternstrings empty? return three groups
            self.compiled_regex_pattern = re.compile(f"({multiple_patternstrings[0]})(.*)({multiple_patternstrings[1]})", rflag)
            self.group_of_interest = 2
            #return (re.compile(f"({multiple_patternstrings[0]})(.*)({multiple_patternstrings[1]})", rflag), 2, 3)
        
    


from pdfminer.high_level import extract_text
from docx import Document
import re
from .patterns import Pattern

class DocumentProcessor:
    text: str
    result: dict[str, str]
    result_required: dict[str, str]
    
    def __init__(self, file_path):
        self.extract_text(file_path = file_path)
    
    def extract_text(self, file_path):
        raise NotImplementedError("This method should be implemented by subclasses.")
    
    # function to process the "terms and pattern group" from ConfigStrategy
    def terms_patterns(self, tap: dict[str, Pattern]):
        result = {}
        result_required = {}
        for term, p in tap.items():
            mo = re.search(p.compiled_regex_pattern, self.text)
            content = None
            has_groups = p.compiled_regex_pattern.groups > 0
            # have patterns been found at all?
            if mo:
                # are groups present?
                if has_groups:
                    #goi = p.group_of_interest
                    #number_of_groups = p.compiled_regex_pattern.groups 
                    content = mo.group(p.group_of_interest)
                    # # in case only two groups present: limit length of matched text
                    # if p.pattern_options.cropToLength:
                    #     if len(content) > p.pattern_options.cropToLength and number_of_groups == 2:
                    #         if goi == 1:
                    #             content = content[-p.pattern_options.cropToLength:]
                    #         else:
                    #             content = content[:p.pattern_options.cropToLength]
                # no groups
                else:
                    # mos: indices of matched text
                    mos = mo.span()
                    content = self.text[mos[0]:mos[1]]
            
            # further process content
            # linebreak to space
            if content and p.pattern_options.lineBreakToSpace:
                content = re.sub("\n", " ", content)
                
            # cropToLength
            if content and p.pattern_options.cropToLength:
                ctl = p.pattern_options.cropToLength + 1 # adding 1 makes sure the length of cropped string equals cropToLength
                #print(f"pattern: {p.compiled_regex_pattern}, content: {content}, ctl: {ctl}, groups: {p.compiled_regex_pattern.groups}, goi: {p.group_of_interest}\n")
                if len(content) > ctl:
                    if p.compiled_regex_pattern.groups == 2 and p.group_of_interest == 1:
                        content = content[-ctl:]
                    else:
                        content = content[:ctl]
            
            # strip
            if content:
                content = content.strip()
                
            # result required
            if p.pattern_options.required:
                # print(f"\nGoing into required - Pattern: {p}\n")
                # print(f"content: {content!r}\n")
                result_required[term] = content
            
            # document result
            #print(f"\nGoing into all: {p}\n")
            result[term] = content
        # put into class attribute
        self.result = result
        self.result_required = result_required
        
        # # Finally 
        # print(f"all results: {self.result!r}\n")
        # print(f"required: {self.result_required!r}\n")
        # na_required = any([n is None for n in self.result_required.values()])
        # na_all = any([n is None for n in self.result.values()])
        # print(f"none in required: {na_required}, none in all: {na_all}\n")
        
    def all_found(self) -> bool:
        return not any([n is None for n in self.result_required.values()])
        
    def terms_content(self, tap) -> dict[str, str]:
        self.terms_patterns(tap)
        return self.result
                
    def contains(self, text_: str) -> bool:
        pattern = re.compile(text_)
        if pattern.search(self.text):
            return True
        else:
            return False
        
class PDFProcessor(DocumentProcessor):
    def extract_text(self, file_path):
        self.text = extract_text(file_path)
        
class TXTProcessor(DocumentProcessor):
    def extract_text(self, file_path):
        with open(file_path, "r") as doc:
            self.text = doc.read()

class DOCXProcessor(DocumentProcessor):
    def extract_text(self, file_path):
        d = Document(file_path)
        txt = ''
        for p in d.paragraphs:
            txt = txt + '\n' + p.text
        self.text = txt
        

# Placeholder for future extensions
# class MarkdownProcessor(DocumentProcessor):
#     ...



import hashlib
import re

class PatternStringError(Exception):
    def __init__(self, msg):
        message_primer = "There appears to be a problem with the patterstring:\n"
        self.message = message_primer + msg
        super().__init__(self.message)

def file_sha256(file_name: str) -> str:
    with open(file_name, "rb") as f:
        bytes = f.read()
        hex_hash = hashlib.sha256(bytes).hexdigest()
        return hex_hash
    
def check_regex(patternstring: str) -> bool:
    try: 
        rgx = re.compile(patternstring)
        # Also, do not allow regex groups
        if rgx.groups > 0:
            return False
    except:
        return False
    return True

# process_terms
# This function has the following jobs:
#   - check if patternstrings can be converted to regex patterns (type re.Pattern)
#   - check if patternstrings already contain a "matchall pattern" (.*), as these are not allowed 
#   - create capture groups if divider is present; if present:
#       - check if divider occurs more than once, as this is not allowed
#       - replace the divider by a capture group matching all ("matchall pattern")
#       - return the index of the (one and only) capture group representing the matchall pattern
#       - return the total number of capture groups
# Return value:
#   The function returns a tuple of (re.Pattern, int, int) containing the compiled pattern,
#   the index of the group containing the (one and only) matchall pattern, and
#   the number of capture groups present.
#   In case no capture groups have been formed, the second and third integers are set to -1.
def process_terms(patternstring: str, divider: str = "@@@", rflag: re.RegexFlag = re.NOFLAG) -> tuple[re.Pattern, int, int]:
    # if patternstrings contains groups, reject
    if not check_regex(patternstring):
        raise PatternStringError(f"Concerning '{patternstring}': cannot be used as regex pattern; also, regex groups are not allowed!")
    # helper to check if pattern only consists of a matchall pattern
    def matchall_only(s) -> bool:
        return re.search("\.\*", s) and len(s) == 2
    # check if matchall pattern is present (as this is not allowed)
    if matchall_only(patternstring):
        raise PatternStringError(msg=f"The string '{patternstring}' only contains the matchall-pattern '.*' and can therefore not be processed.")
    # divider_hits counts the number of divider in the string; only one is allowed (see below)
    divider_hits = len(re.findall(divider, patternstring))
    # check the number of occurrences of divider
    if divider_hits > 1:
        # as this is not implemented, throw an error
        raise PatternStringError(msg=f"Each term must correspond to either *one regex pattern* or *two regex patterns divided by '{divider}'*!")
    if divider_hits == 0:
        # return without capture groups
        return (re.compile(patternstring, rflag), -1, -1)
    # process the patternstrings divided by divider
    multiple_patternstrings = re.split(divider, patternstring)
    
    # check if patternstring and multiple_patternstrings are valid regex patterns without groups
    all_strings = multiple_patternstrings.copy()
    all_strings.append(patternstring)
    all_check = [check_regex(s) for s in all_strings]
    if not all(all_check):
        raise PatternStringError(f"Concerning one of '{all_strings!r}': cannot be used as regex pattern; also regex groups are not allowed!")
    
    # do any of the patternstrings only contain a matchall pattern?
    if any([matchall_only(p) for p in multiple_patternstrings]):
        raise PatternStringError(msg=f"At least one of '{multiple_patternstrings!r}' only consists of a matchall-pattern '.*' and can therefore not be processed.")
    # is any of the patternstrings of length 0?
    lenx = [len(i) for i in multiple_patternstrings]
    lenx0 = [l == 0 for l in lenx]
    # if yes
    if any(lenx0):
        # the first?
        if lenx0[0]:
            return (re.compile(f"(.*)({multiple_patternstrings[1]})", rflag), 1, 2)
        # the second?
        return (re.compile(f"({multiple_patternstrings[0]})(.*)", rflag), 2, 2)
    # none of the patternstrings empty? return three groups
    return (re.compile(f"({multiple_patternstrings[0]})(.*)({multiple_patternstrings[1]})", rflag), 2, 3)


from typing import List
from sqlalchemy import ForeignKey
from time import time_ns
from sqlalchemy import String, CHAR, create_engine
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import mapped_column
from sqlalchemy.orm import relationship

class Base(DeclarativeBase):
    pass

class DbRun(Base):
    __tablename__ = "run"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    title: Mapped[str] = mapped_column(String(40))
    yml_filename: Mapped[str] = mapped_column(String(40))
    yml_sha256: Mapped[str] = mapped_column(String(64))
    timepoint: Mapped[int] = mapped_column(default=time_ns())
    strategies: Mapped[List["DbStrategy"]] = relationship(back_populates="run", cascade="all, delete-orphan")
    def __repr__(self) -> str:
        return f"Run(id={self.id!r}, title={self.title!r}, yml_sha256={self.yml_sha256!r})"

class DbStrategy(Base):
    __tablename__ = "strategy"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    run_id: Mapped[int] = mapped_column(ForeignKey("run.id"))
    name: Mapped[str] = mapped_column(String(100))
    processed_directory: Mapped[str] = mapped_column(String(200))
    file_name_pattern: Mapped[str] = mapped_column(String(100))
    file_content_pattern: Mapped[str] = mapped_column(String(100), nullable=True)
    export_format: Mapped[str] = mapped_column(String(10), nullable=True)
    export_path: Mapped[str] = mapped_column(String(200))
    export_csv_divider: Mapped[str] = mapped_column(CHAR, nullable=True)
    run: Mapped["DbRun"] = relationship(back_populates="strategies")
    processed_files: Mapped[List["DbProcessedFile"]] = relationship(back_populates="strategy", cascade="all, delete-orphan")
    def __repr__(self) -> str:
        return f"Strategy(id={self.id!r}, export_path={self.export_path!r})"
    
class DbProcessedFile(Base):
    __tablename__ = "processed_file"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    strategy_id: Mapped[int] = mapped_column(ForeignKey("strategy.id"))
    filename: Mapped[str] = mapped_column(String(40))
    sha256: Mapped[str] = mapped_column(String(64))
    strategy: Mapped["DbStrategy"] = relationship(back_populates="processed_files")
    matches: Mapped[List["DbMatch"]] = relationship(back_populates="processed_file")
    def __repr__(self) -> str:
        return f"File(id={self.id!r}, name={self.name!r}, sha256={self.sha256!r})"
    
class DbMatch(Base):
    __tablename__ = "match"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    file_id: Mapped[int] = mapped_column(ForeignKey("processed_file.id"))
    term: Mapped[str] = mapped_column(String(30))
    content: Mapped[str] = mapped_column(String(500), nullable = True)
    processed_file: Mapped["DbProcessedFile"] = relationship(back_populates="matches")
    def __repr__(self) -> str:
        return f"Match(id={self.id!r}, term={self.term!r}, content={self.content!r})"
    